# -*- coding: utf-8 -*-
"""
Created on Fri May 14 23:09:01 2021

@author: ErnestoAngelCurbeloB
"""
from proposal import Prop

import numpy as np

from numpy.matlib import repmat

from scipy.integrate import quad,dblquad

from scipy.stats import multivariate_normal as mvn

from time import time

import matplotlib.pyplot as plt
plt.style.use("default")

class Ejemplo:
    """THIS IS THE FIRST EXAMPLE OF THE PAPER,
    THE DIMENSION OF THE PARAMETRES IS ONE"""
    
    # INITIALIZE THE CLASS
    def __init__(self,xtrue,sigmatrue,K,sigma):
        
        self.K = K
        self.xtrue = xtrue
        self.sigmatrue = sigmatrue
        # self.yk = self.f(xtrue) + mvn.rvs(cov=(sigmatrue**2)*np.eye(K),size=1)
        self.yk = np.loadtxt("data_y.txt")
        self.sigma = sigma
    
    # THE FUNCTION OF THE OBSERVATIONAL MODEL
    def f(self,x):
        
        try:

            n = len(x)

            a = x**2 + np.log(abs(np.sin(10*x)))
            
            ff = np.ones(n)
            
            ff=ff*a
            
            ff = repmat(np.asmatrix(ff).T,1,self.K)
            
        except:
            a = x**2 +  np.log(abs(np.sin(5*x)))
            
            ff = np.ones(self.K)*a
        
        return ff
    
    # LIKELIHHOD OF THE MODEL
    def likelihood(self,x):
        
        normal = mvn(mean=self.yk,cov=(self.sigma**2)*np.eye(self.K))
        
        return normal.pdf(self.f(x))
        # return normal.logpdf(self.f(x))
    
    # PRIOR DENSITY FOR X
    def prior_x (self,x):
        
        return 1*(x>=0)*(x<=20)/20
    
    # PRIOR DENSITY FOR SIGMA
    def prior_sigma(self,sig):
        
        return 1*(sig>=0)*(sig<=20)/20
    
    # CONDITIONAL POSTERIOR FOR X
    def post_x(self,x):
        
        return self.likelihood(x)*self.prior_x(x)
    
    # COMPLETE POSTERIOR
    def post_all(self,x,sig):
        """This is not a density, it does not itegrate 1"""
        
        self.sigma = sig
        
        a = self.likelihood(x)*self.prior_x(x)*self.prior_sigma(sig)
        
        # def y(x,sig):
        #     self.sigma = sig
        #     return self.likelihood(x)*self.prior_x(x)*self.prior_sigma(sig)
        # Z = dblquad(y, 0.01, 20, lambda sig:0.01,lambda sig:20)[0]
        
        return a#/Z
    
    def plot_post(self):
        
        X = np.arange(0.1,4,0.01)
        
        Sig = np.arange(0.1,10,0.01)
        
        aux_post = lambda sig,x: self.post_all(x, sig)
        
        int_ = dblquad(aux_post,0,20,lambda sig:0,lambda sig:20)[0]
        
        print(int_)
        
        Z = np.zeros((len(Sig),len(X)))
        
        for si in range(len(Sig)):
            
            for xi in range(len(X)):
                
                Z[si,xi] = self.post_all(X[xi], Sig[si])
        
        X,Sig = np.meshgrid(X,Sig)
        
        fig = plt.figure(figsize=(10,6))
        
        ax = fig.add_subplot(111)

        ax.imshow(Z/int_,extent=[0,4,0,10],origin="lower",aspect="auto",alpha=1)
        
        ax.grid(False)

        
        return #fig
                
    
    def plot_post_x(self,sigm = None,xlim=(0.0001,8)):
        
        self.sigma = self.sigmatrue
        
        if sigm is not None:
            
            old_sigm = self.sigma
            
            self.sigma = sigm
        
        
        
        print(self.sigma)
                
        Z = quad(self.post_x,0.1,20)
    
        x = np.arange(xlim[0],xlim[1],0.001)
   
        plt.plot(x,self.post_x(x)/Z[0],color = "orangered")
        
        return #fig
    
    def marginal_post_x(self,x):
        
        y = lambda sig: self.post_all(x,sig)
        
        return quad(y, 0.01, 20)[0]
    
    def marginal_post_sig(self,sig):
        
        y = lambda x: self.post_all(x,sig)
        
        return quad(y,0.01,20)[0]
    
    # FUNCTIONS TO PLOT
    def plot_marginal_post_x(self,xlim=(0.01,8)):
        
        x = np.arange(xlim[0],xlim[1],0.01 )
        
        Z = quad(self.marginal_post_x,0.00001,10)[0]
        
        y = np.array([self.marginal_post_x(i)/Z for i in x.tolist()])
        
        plt.plot(x,y)
        
        return #fig
    
    def plot_marginal_post_sig(self,siglim=(0.01,20)):
        
        sig = np.arange(siglim[0],siglim[1],0.01)
        
        Z = quad(self.marginal_post_sig,0.00001,20)[0]
        
        y = np.array([self.marginal_post_sig(i)/Z for i in sig.tolist()])
        
        fig = plt.figure()
        
        ax = fig.add_subplot(111)
        
        ax.plot(sig,y)
        
        ax.grid(False)
        
        return #fig
    
    def plot_f(self):
        
        # plt.plot(self.yk,"*")
        
        x = np.arange(0.001,20,0.1)
        
        y = [self.f(xi) for xi in x.tolist()]
        
        # plt.plot([0,7],[self.f(self.xtrue),self.f(self.xtrue)])
        plt.plot(x,y)
        
        return
    # ========================================================================

    # SAMPLING WITH ATAIS
    def new_sampling(self,initial_value_mu,initial_value_sigma,sigma_prop_init,T,N,temporal = "No"):
    
        pi_map = 0
        
        sigma_prop = sigma_prop_init
        
        self.sigma = initial_value_sigma
        
        sigmas = [] # keep track of all the sigmas
        
        mu_t = initial_value_mu
        
        all_weights = np.empty((T,N))
        
        all_weights_normalized = np.empty((T,N))
        
        all_samples = np.empty((T,N))
        
        e_n_t_s =  np.zeros((T,N))
        
        sum_of_weights = np.zeros(T)
        
        sigma_ml = initial_value_sigma
        
        sigmas_ml = [sigma_ml]
        
        xs_map = []
        
        # auxiliar_evaluations_exp = np.zeros((T,N))
        
        auxiliar_evaluations_proposals = np.zeros((T,N))
        
        # all_priors = np.zeros((T,N))
        
        x_map = initial_value_mu
        
        for t in range(0,T):
            #SAMPLING ###################################

            proposal = mvn(mu_t,sigma_prop) #DEFINE PROPOSAL
            
            sample = proposal.rvs(N)

            evaluation_target_t = self.post_x(sample)

            evaluation_proposal_t = proposal.pdf(sample) # EVALUATE PROPOSAL

            e_n_t = np.linalg.norm(self.yk - self.f(sample),axis=1)**2

            e_n_t_s[t,:] = e_n_t

            weights_t = evaluation_target_t/evaluation_proposal_t # CALCULATE WEIGHTS
            
            weights_t = weights_t + 1e-231
            ##############################################
            sum_ = (weights_t.sum())

            weights_t_normalized = weights_t/(sum_) #NORMALIZE THE WEIGHTS
            
            auxiliar_evaluations_proposals[t,:] = evaluation_proposal_t#*sum_
            
            all_weights[t,:] = weights_t # SAVE THIS WEIGHT
            
            all_weights_normalized[t,:] = weights_t_normalized # SAVE NORMALIZED WEIGHTS
                        
            all_samples[t,:] = sample #SAVE THE SAMPLES
            
            sum_of_weights[t] = sum_
            
            argmax_i = np.argmax(evaluation_target_t)

            x_hat_t = sample[argmax_i] 

            evaluation_target_max_t = evaluation_target_t[argmax_i]
            
            sig_hat_t = np.linalg.norm(self.yk - self.f(x_hat_t))/(np.sqrt(self.K))
                       
            # GLOBAL MAXIMUM ESTIMATION#####################################
            if sig_hat_t < sigma_ml:
            
                sigma_ml = sig_hat_t
            
            if evaluation_target_max_t > pi_map:
                
                pi_map = evaluation_target_max_t
                
                x_map = x_hat_t
                
            ##################################################################
            
            self.sigma = sigma_ml
            
            sigmas_ml.append(sigma_ml)
            
            xs_map.append( x_map )
            
            sigmas.append(self.sigma)
            

            # ADAPTATION #####################################################
            # print(sample)
            x_bar = (weights_t_normalized * sample).sum()
            
            mu_t = x_map
            # sigma_prop = (((sample-x_bar)**2)*weights_t_normalized).sum()
            sigma_prop = 7
            
            #####################################################
            # print(t)
        #END FOR
        
        
        # UPDATE THE WEIGHTS 
        # den = ((2*np.pi)**self.K)*np.array(sigmas_ml[:-1])
        
        # exp_ = np.exp(-auxiliar_evaluations_exp)
        
        # den2 = repmat(np.matrix(den).T,1,N)
        
        # lk = np.array(den2) * (exp_) * all_priors
        
        # print(lk,sep="\n\n")
        
        evaluation_last_pi = np.zeros((T,N))
        
        if temporal == "yes":
            # print("Updating the weights usin temporal denominator")
            evaluation_proposal_temporal = np.zeros((T,N))
            
            for t in range(T):
                
                evaluation_last_pi[t,:] = self.post_x(all_samples[t,:,:])
                
                # evaluation_proposal_temporal[t,:] = 0
                
                for n in range(N):
                    # print("bug",(all_samples[t,:,:]))
                    # b =  mvn(mean = xs_map[t],cov=sigma_prop).pdf(all_samples[t,:,:])
                    # print("bug2",b)
                    evaluation_proposal_temporal[t,:] += mvn(mean = xs_map[t],cov=sigma_prop).pdf(all_samples[t,:,:])
                    
                    
            updated_weights = evaluation_last_pi/evaluation_proposal_temporal
                
                # flatted_all_samples = all_samples.ravel()
        
        else:
            # print("\nUpdating the weights using usual denominator\n")
            for t in range(T):
                # print(f"\n Iteración {t}")
                evaluation_last_pi[t,:] = self.post_x(all_samples[t,:])
            # print("\n Doing the cocient")
            # print(evaluation_last_pi.shape)
            updated_weights = evaluation_last_pi/auxiliar_evaluations_proposals
            # print("finish the cocient")
        # print(evaluation_last_pi)
        # print("\nSaving the attributes\n")
        self.all_samples = all_samples
        
        self.updated_weights = updated_weights
        
        self.sigmas_ml = sigmas_ml
        
        self.e_n_t_s = e_n_t_s
        
        self.auxiliar_evaluations_proposals = auxiliar_evaluations_proposals
        
        # print("Finish saving the attributes")
    
        
    # return self.all_samples,self.updated_weights,self.sigmas_ml,self.e_n_t_s,self.auxiliar_evaluations_proposals,xs_map
        
    def plot_sigmas_ml(self):
        
        try:
        
            if self.sigmas_ml is None:
                
                print("First execute sampling method")
                
                return
        except AttributeError:
            
            print("\nFirst execute sampling method")
                
            return
        else: 
            
            fig = plt.figure()
            
            ax = fig.add_subplot(111)
            
            ax.plot(self.sigmas_ml,label="Sigmas ml")
            
            plt.legend()
            
            plt.show()

            return                

    def rho_n_t_sigma(self,sigma):
        
        try:
            
            self.all_samples
            
            self.auxiliar_evaluations_proposals
            
        except:
            
            print("\n First execute sampling method")
            
            return
        
        
        t1 = time()
        
        N = self.all_samples.shape[1]
        
        T = self.all_samples.shape[0]
        
        s1 = (2*np.pi*sigma**2)**(self.K/2)

        s2 = np.exp(-self.e_n_t_s/(2*sigma**2))
        
        s3 = np.apply_along_axis(self.prior_x,arr=self.all_samples,axis=1)

        s4 = self.auxiliar_evaluations_proposals
                
        s = s2*s3/(s1*s4)
        
        ans = s.sum()/(N*T)
        
        
        return ans,s
    
    def joint_posterior(self,J,n_samples_per_sigma,plot = True):
        
        """On this function the expected value of x and sigma is calculated
        The output of the code is (E(x),E(sigma))
        """
        
        T = self.all_samples.shape[0]
        
        N = self.all_samples.shape[1]
        
        linf = 0.001
        
        lsup = 20
        
        npoints = 1000

        test = Prop(self.marginal_post_sig,linf,lsup,npoints)

        sigmas = [test.sampling()[0] for i in range(1000)]
        
        h = lambda x,sig: np.array([x, sig])
        
        hist_x = np.empty(J*n_samples_per_sigma)
        
        hist_sigma = np.empty(J*n_samples_per_sigma)
        
        rhos_per_sigma = []
        
        ans = 0
        
        for j in range(J):
            
            aux = self.rho_n_t_sigma(sigmas[j])[1]
            
            aux_sum = aux.sum()
            
            Aux = self.all_samples * aux/aux_sum 
            
            ans += Aux.sum()
            
            flatted_aux =(aux/aux_sum).ravel()
            
            choosen_indices = np.random.choice(np.arange(N*T),p=flatted_aux,size=n_samples_per_sigma)
            
            hist_x[j*n_samples_per_sigma:(j+1)*n_samples_per_sigma] = self.all_samples.ravel()[choosen_indices]
            
            hist_sigma[j*n_samples_per_sigma:(j+1)*n_samples_per_sigma] = sigmas[j]
            
            rhos_per_sigma.append(aux)
        
        if plot == True:
            
            
            plt.figure()
            plt.scatter(hist_x,hist_sigma,s=1)
            
            plt.figure()
            self.plot_post()
            
            plt.figure()
            n_ = int(np.sqrt(J*n_samples_per_sigma))
            hist = plt.hist2d(hist_x,hist_sigma,bins=n_)
            # plt.imshow(hist[0],extent=(0,max(hist_x),0,max(hist_sigma)),aspect="auto",origin="lower")
            # print(hist)
            
            plt.figure()
            n_ = int(np.sqrt(len(hist_x)))
            plt.hist(hist_x,bins=n_,density=True)
            self.plot_marginal_post_x()
            plt.show()
            
        return ans/J,np.mean(sigmas)
    
    
    
        
    def Z_hat(self,R):

        sigma_arr = np.linspace(0.1,20,R)
        
        z_hat = [self.rho_n_t_sigma(s)[0] for s in sigma_arr.tolist()]
        
        z_hat = sum(z_hat)/R
        
        return z_hat
    
    # def hist_resample_samples(self):
        
    #     W = self.updated_weights.ravel(order="F")
        
    #     flatted_sample = self.all_samples.ravel(order="F")
        
    #     W_normalized = W/(W.sum())
        
    #     resample =np.random.choice(np.arange(len(flatted_sample)),p=W_normalized,
    #                                size=100000)
        
    #     plt.hist(flatted_sample[resample],bins=1000,density=True)
        
    #     self.plot_post_x()
        
    #     plt.show()

    #     return
    
# FUNCTIONS FOR CALCULATE THE TRUE VALUES     
    def grid_post_all(self,step_x,step_sigma):
        
        x_s = np.arange(0.001,20,step_x)
        
        sigma_s = np.arange(0.001,20,step_sigma)
        
        grid = 0
        
        for x in x_s:
            
            for sig in sigma_s:
                
                grid += step_x*step_sigma*self.post_all(x,sig)        
                
        return grid
        
    def grid_marginal_post_x(self,step=0.1):
        
        x = np.arange(0.001,20,step)
        
        grid = 0
        
        for x_ in x.tolist():
            
            grid += step*self.marginal_post_x(x_)
            
        return grid
    
    def grid_marginal_post_sigma(self,step):
        
        sig = np.arange(0.001,20,step)
        
        grid = 0
        
        for s in sig.tolist():
            
            grid += step*self.marginal_post_sig(s)
            
        return grid
    
    def true_expected_value_x(self,step=0.1):
        
        expected = lambda x: x*self.marginal_post_x(x)
        
        x = np.arange(0.001,20,step)
        
        Z = 2.9603586442501607e-12
        
        expected_value = 0
        
        for x_ in x.tolist():
            
            expected_value += step*expected(x_)
        
        return expected_value/Z
    
    def true_expected_value_sigma(self,step=0.1):
        
        expected = lambda sig: sig*self.marginal_post_sig(sig)
        
        sigmas = np.arange(0.001,20,step)
        
        Z = 2.65509328242642e-12

        
        expected_value = 0
        
        for sig in sigmas.tolist():
            
            expected_value += step*expected(sig)
            
        return expected_value/Z
    # ========================================================================

    def hist_joint(self):
        
        linf = 0.001
        
        lsup = 20
        
        npoints = 1000

        # print("\n building test")
        test = Prop(self.marginal_post_sig,linf,lsup,npoints)
        # print("\n sigmaaaaas")
        sigmas_samples = [test.sampling()[0] for i in range(5000)]
        bin_ = int(np.sqrt(len(sigmas_samples)))
        
        x_to_hist = []
        # print("\n going to sample")
        for sigma_j in sigmas_samples:
            
            # print("\n",sigmas_samples.index(sigma_j))
            
            rhos = self.rho_n_t_sigma(sigma_j)[1]
            
            flatted_all_samples = self.all_samples.ravel()
            
            flatted_rhos = rhos.ravel()
            
            p = flatted_rhos/flatted_rhos.sum()
            
            a = np.random.choice(np.arange(len(p)),p=p)
            
            x_to_hist.append(flatted_all_samples[a])
            
        # print("\n finish sampling")
        plt.scatter(x_to_hist,sigmas_samples,s=1)
        
        plt.figure()
        plt.hist2d(x_to_hist,sigmas_samples,bins=bin_)
        
        plt.figure()
        plt.hist(x_to_hist)
        
        return 
            
    def conditional_x(self,sigma):
        
        rhos = self.rho_n_t_sigma(sigma)[1]
        
        self.sigma = sigma
        
        x = np.arange(0.001,5,0.01)
        
        Z = quad(self.post_x,0.001,20)[0]
        
        plt.plot(x,self.post_x(x)/Z)
        
        normed_rhos = rhos/rhos.sum()
        
        f_normed_rhos = normed_rhos.ravel()
        
        ind = np.random.choice(np.arange(len(f_normed_rhos)),p=f_normed_rhos,size=1000000)
        
        f_samples = self.all_samples.ravel()
        
        bin_ = int(np.sqrt(len(f_normed_rhos)))
        
        plt.hist(f_samples[ind],bins=bin_,density=True)
        
        return

    
   
        
        
    
# K = 8
# xtrue = 2.5
# sigmatrue = 4
# sigma = 4
# ex = Ejemplo( xtrue,sigmatrue,K,sigma)


# N = 10000
# T = 100
# initial_value_mu = 5
# initial_value_sigma = 20
# sigma_prop_init = 1

# print(ex.new_sampling(initial_value_mu, initial_value_sigma, sigma_prop_init, T, N))
# print("\n To the joint posterior")
# print(ex.joint_posterior(1000,n_samples_per_sigma=100))

# ex.hist_samples()
# ex.hist_resample_samples()
# print("\n",ex.rho_n_t_sigma(1))
# ex.Z_hat(5000)
# print("\n",ex.sigma)
# ex.plot_marginal_post_x()
# ex.plot_marginal_post_sig()
# ex.plot_post()

# print(ex.grid(0.01,0.01))

# print(ex.grid_marginal_post_x(0.001))

# print(ex.grid_marginal_post_sigma(0.001))

# print(ex.true_expected_value_x(0.001))

# print(ex.true_expected_value_sigma(0.001))


#Z con grid 3.0325054711079082e-12
# Z con quad 2.608523330382332e-12
# Z con grid más finos 2.9662643134043414e-12

# Z de marginal posterior de x 2.9669000867956394e-12
# Z de marginal posterior de x con grid más fino 2.9603586442501607e-12

# Z de marginal posterior de sigma 2.6550964982152915e-12   2.655092505837558e-12
# Z de marginal posterior de sigma con grid más fino 2.655092505837558e-12
# Z de marginal posterior de sigma con grid más fino 2.65509328242642e-12

# valor esperado de x 1.143042553262111
# valor esperado de x con grid mas fino 1.1057887665488222
# valor esperado de x con grid mas fino 1.1007418448964839

# valor esperado de sigma  4.8215735046107
# valor esperado de sigma con grid más fino 4.821585972977968
# valor esperado de sigma con grid más fino aún 4.821591040494352

### SIGMA ML = 


# linf = 0.001
# lsup = 20
# npoints = 50
# n_samples = 1000
# initial_point = 3
# cov = 1
# test = Prop(ex.marginal_post_sig,linf,lsup,npoints)
# test.plot()
# l = [test.sampling()[0] for i in range(1000)]
# plt.hist(l,bins=50)


















#FORGOTTEN METHOD


# def sampling(self,initial_value_mu,initial_value_sigma,sigma_prop_init,T,N):
        
#         pi_map = 0
        
#         sigma_prop = sigma_prop_init
        
#         self.sigma = initial_value_sigma
        
#         sigmas = [] # keep track of all the sigmas
        
#         mu_t = initial_value_mu
        
#         all_weights = np.empty((T,N))
        
#         all_weights_normalized = np.empty((T,N))
        
#         all_samples = np.empty((T,N))
        
#         sigma_ml = initial_value_sigma
        
#         sigmas_ml = [sigma_ml]
        
#         xs_map = []
        
#         x_map = None
        
#         for t in range(0,T):
#             #SAMPLING ###################################
#             proposal = mvn(mu_t,sigma_prop) #DEFINE PROPOSAL
            
#             sample = proposal.rvs(N) #GET SAMPLES
            
#             evaluation_target_t = self.post_x(sample) #EVALUATE TARGET
#             return sample.shape, evaluation_target_t.shape
#             evaluation_proposal_t = proposal.pdf(sample) # EVALUATE PROPOSAL
            
#             return evaluation_proposal_t
#             weights_t = evaluation_target_t/evaluation_proposal_t #CALCULATE WEIGHT
#             ##############################################

#             weights_t_normalized = weights_t/(weights_t.sum())
            
#             all_weights[t,:] = weights_t
            
#             all_weights_normalized[t,:] = weights_t_normalized
            
#             all_samples[t,:] = sample
            
#             argmax_i = np.argmax(evaluation_target_t)
            
#             x_hat_t = sample[argmax_i] 
            
#             evaluation_target_max_t = evaluation_target_t[argmax_i]
            
#             sig_hat_t = np.linalg.norm(self.yk - self.f(x_hat_t))/(np.sqrt(self.K))
            
#             if sig_hat_t < sigma_ml:
            
#                 sigma_ml = sig_hat_t
            
#             if evaluation_target_max_t > pi_map:
                
#                 pi_map = evaluation_target_max_t
                
#                 x_map = x_hat_t
            
#             sigmas_ml.append(sigma_ml)
            
#             xs_map.append( x_map )
            
#             sigmas.append(self.sigma)
            
            
#             # Update the values of the proposal
            
#             mu_t = x_map
            
#             x_bar = (sample*weights_t_normalized).sum()
            
#             sigma_prop = (((sample - x_bar)**2)*weights_t_normalized).sum() + 0.1
            
#         return all_samples,all_weights_normalized  


        
    # def hist_samples(self):
        
    #     a = self.all_samples.ravel()
        
    #     sqr = int(np.sqrt(len(a)))
        
    #     plt.hist(a,bins=sqr,density =True)
        
    #     plt.show()
        
    #     return